from omigo_core import utils, timefuncs
import datetime

# helper function to add relative time option in splunk and logscale
def resolve_time_str(x):
    # check for specific syntax with now
    if (x.startswith("now")):
        x = x.replace(" ", "")
        base_time = datetime.datetime.utcnow().replace(tzinfo = datetime.timezone.utc)
        diff_sec = None
        # check if there are any diff units
        if (x == "now"):
            diff_sec = 0
        else:
            # validation for parsing logic
            if (x.startswith("now-") == False):
                raise Exception("Unknown operator against now: ", x)

            # take the diff part
            diffstr = x[len("now-"):]

            # unit is single letter, 'd', 'h', 'm', 's'
            unit = diffstr[-1]

            # how many of units to apply
            count = int(diffstr[0:-1])
            if (unit == "d"):
                diff_sec = count * 86400
            elif (unit == "h"):
                diff_sec = count * 3600
            elif (unit == "m"):
                diff_sec = count * 60
            elif (unit == "s"):
                diff_sec = count * 1
            else:
                raise Exception("Unknown time unit:", parts[1])

        # return base_time minus diff
        return timefuncs.utctimestamp_to_datetime_str(int(base_time.timestamp()) - diff_sec)
    else:
        return timefuncs.utctimestamp_to_datetime_str(timefuncs.datetime_to_utctimestamp_sec(x))




from omigo_core import utils
from scipy import stats

# This is WIP
def do_ttest(xtsv1, xtsv2, f, n = 10000, alpha = 0.01):
    xtsv1 = xtsv1.select(f)
    xtsv2 = xtsv2.select(f)

    # find the value of n
    effective_n = min(n, xtsv1.num_rows(), xtsv2.num_rows())

    # perform t-test
    x_pos_1 = xtsv1.sample_n(effective_n)
    x_neg_1 = xtsv2.sample_n(effective_n)

    f_pos = x_pos_1.col_as_float_array(f)
    f_neg = x_neg_1.col_as_float_array(f)

    # sample may not be exact
    effective_n = min(len(f_pos), len(f_neg))
    f_pos = f_pos[0:effective_n]
    f_neg = f_neg[0:effective_n]

    # check for alpha
    if (alpha > 0):
        alpha_2_int = int(effective_n * alpha / 2)
        f_pos = f_pos[alpha_2_int:effective_n - alpha_2_int]
        f_neg = f_neg[alpha_2_int:effective_n - alpha_2_int]

    ttest = stats.ttest_ind(f_pos, f_neg, equal_var = False)
    kstest = stats.ks_2samp(f_pos, f_neg)

    mean_pos = sum(f_pos) / len(f_pos)
    mean_neg = sum(f_neg) / len(f_neg)

    print("Org Len: xtsv1: {}, xtsv2: {}".format(xtsv1.num_rows(), xtsv2.num_rows()))
    print("New Len: xtsv1: {}, xtsv2: {}".format(len(f_pos), len(f_neg)))
    print("t-test: mean: xtsv1: {}, xtsv2: {}, pvalue: {}, statistic: {}".format(mean_pos, mean_neg, ttest.pvalue, ttest.statistic))
    print("ks-test pvalue: {}, ks-test statistic: {}, location: {}, sign: {}".format(kstest.pvalue, kstest.statistic, kstest.statistic_location, kstest.statistic_sign))




import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio
pio.renderers.default = "notebook"

# Cycle through a palette if you need more than 26
def __get_sankey_colors__(n, palette = "Vivid"):
    base_colors = getattr(px.colors.qualitative, palette)
    # Repeat palette if needed
    return [base_colors[i % len(base_colors)] for i in range(n)]


def get_sankey_figure(mp, pad = 15, thickness = 20, line_color = "black", line_width = 0.5, title_text = "Basic Sankey", font_size = 10):
    # Define nodes and links
    fig = go.Figure(data=[go.Sankey(
        node = dict(
          pad = pad,
          thickness = thickness,
          line = dict(color = line_color, width = line_width),
          label = mp["label"],
          color = __get_sankey_colors__(len(mp["label"]))
        ),
        link = dict(
          source = mp["source"],
          target = mp["target"],
          value = mp["value"]
      ))])

    # update layout
    fig.update_layout(title_text = title_text, font_size = font_size)

    # return
    return fig





from omigo_core import dataframe, utils, timefuncs, funclib
import json

class HadoopSqlBase:
    def __init__(self):
        pass

    def execute_query(self, columns = ["*"], table = None, where_clause = "", group_by_cols = None, having_clause = "",
        order_by_cols = None, sort_order = None, distinct_flag = None, limit = None, map_col = None):

        # warn
        utils.warn_once("HadoopSqlBase: execute_query: this is provided as reference implementation and should be used carefully in prod to avoid issues with sql")

        # some validation
        if (table is None):
            raise Exception("HadoopSqlBase: execute_query: table can not be none")

        # check for columns. TODO: read the default set of columns from table
        if ("*" in columns):
            utils.warn_once("HadoopSqlBase: execute_query: select * is not fully supported")

        # create effective_columns
        effective_columns = []

        # for group by queries, add the group by columns to the select columns
        if (group_by_cols is not None):
            for gcol in group_by_cols:
                found = False
                for col in columns:
                    col = col.lower()
                    index = col.find(" as ")
                    # check for aliases
                    if (index != -1):
                        col = col[0:index].strip()

                    # check for match
                    if (col == gcol.lower()):
                        found = True
                        break

                # if col is not found, add to effective_columns
                if (found == False):
                    effective_columns.append(gcol)

        # add select colummns after group by
        for col in columns:
            effective_columns.append(col)

        # use empty arrays as defaults
        map_cols = []
        if (map_col is not None):
            map_cols.append(map_col)

        # do a lowercase on mapcols
        utils.warn_once("sql_helper: doing lowercase on map_col: {}".format(map_col))
        map_cols = list([v.lower() for v in map_cols])

        # base query
        if (distinct_flag is None or distinct_flag == False):
            query = "select {} from {}".format(", ".join(effective_columns), table) # nosec
        else:
            query = "select distinct {} from {}".format(", ".join(effective_columns), table) # nosec

        # where clause
        if (where_clause != ""):
            query = "{} where {}".format(query, where_clause)

        # group by clause
        if (group_by_cols is not None and len(group_by_cols) > 0):
            group_by_cols_expr = ",".join(["{}".format(i+1) for i in range(len(group_by_cols))])
            query = "{} group by {}".format(query, group_by_cols_expr)

        # having clause
        if (having_clause != ""):
            query = "{} having {}".format(query, having_clause)

        # order by
        if (order_by_cols is not None):
            query = "{} order by {}".format(query, ", ".join(order_by_cols))

            # check sort order as asc or desc
            if (sort_order is not None):
                query = "{} {}".format(query, sort_order)

        # limit
        if (limit is not None):
            query = "{} limit {}".format(query, limit)

        # debug
        utils.info("HadoopSqlBase: execute_query: {}".format(query))

        # execute query
        start_ts = timefuncs.get_utctimestamp_sec()
        result_cols, result_rows = self.execute_query_in_engine(query)
        end_ts = timefuncs.get_utctimestamp_sec()

        # create the column aliases
        output_cols = []
        map_cols_indexes = {}

        # iterate
        for i in range(len(result_cols)):
            # read column name
            c = result_cols[i]

            # check for map cols. case insensitive match
            if (c in map_cols or c.lower() in map_cols):
                map_cols_indexes[i] = 1

            # add column
            output_cols.append(c)

        # time taken
        time_taken = funclib.get_display_relative_time_str(end_ts - start_ts)

        # debug
        utils.info("HadoopSqlBase: execute_query: num rows: {}, time taken: {}".format(len(result_rows), time_taken))

        # create data
        data_fields = []
        for row in result_rows:
            # create result
            result_row = []

            # row is an array of values
            for i in range(len(row)):
                # read value
                value = row[i]

                # replace None with empty string
                if (value is None):
                    value = ""

                # check if value needs to be converted to serialized json
                if (i in map_cols_indexes.keys()):
                    value = json.dumps(value)

                # utils.info("row: {}".format(value))
                # replace any Ctrl-M characters
                # value = "{}".format(value).replace("\t", " ").replace("\v", " ").replace("\r", " ").replace("\n", " ")

                # append to result
                result_row.append(value)

            # TODO: there can be tabs in the data that can affect serialization
            # cols_str = ["{}".format(t) for t in result_row]
            data_fields.append(result_row)

        # create tsv. Do a validation as this is an external source
        xdf = dataframe \
            .new_with_cols(output_cols, data_fields = data_fields) \
            .validate()

        # return
        return xdf

    def execute_query_in_engine(self, query):
        raise Exception("HadoopSqlBase: execute_query: derived class must implement this method")




from omigo_core import dataframe
from omigo_core import utils
import random
import time
from kafka import KafkaConsumer

# some information that might help
# pip uninstall snappy python-snappy
# pip install python-snappy
# https://github.com/dask/fastparquet/issues/459

# Work In Progress
class KafkaClient:
    # constructor. group_id is created uniquely if not specified.
    def __init__(self, topic, bootstrap_servers, group_id, auto_offset_reset = "latest", value_deserializer = None, excluded_cols = None,
        nested_cols = None):

        # check for value_deserializer
        if (value_deserializer is None):
            value_deserializer = lambda x: x.decode()

        # create consumer
        self.consumer = KafkaConsumer(
            topic,
            bootstrap_servers = bootstrap_servers,
            group_id = group_id,
            auto_offset_reset = auto_offset_reset,
            value_deserializer = value_deserializer
        )

        # json parsing need some tuning parameters
        self.excluded_cols = excluded_cols
        self.nested_cols = nested_cols

        # utils.warn("This KafkaConsumer is Work in Progress.")
        raise Exception("This KafkaClient is not ready")

    # method to read n messages from the topic.
    def read(self, n = 0, max_duration_sec = 0, sampling_rate = 1.0, seed = 0):
        # validtion. either n or max_duration_sec must be non zero
        if (n <= 0 and max_duration_sec <= 0):
            raise Exception("Either n or max_duration_sec must be non zero")

        # initialize random number generator
        random.seed(seed) #nosec

        # initialize start time. this will need better implementation to prevent infinite waiting. TODO
        ts_start = time.time()

        # create header and data fields
        internal_prefix = "__KafkaClient_read__"
        new_header_fields = [internal_prefix]
        new_data_fields = []

        # iterate
        for message in self.consumer:
            # apply sampling
            if (sampling_rate > 1 or random.random() <= sampling_rate):  # nosec
                new_data_fields.append([message.value])

            # check if all messages have been received, or time interval has been reached
            if (n > 0 and len(new_data) >= n):
                break

            # check if max_duration is exceeded
            ts_end = time.time()
            if (max_duration_sec > 0 and int(ts_end - ts_start) >= max_duration_sec):
                break

        # convert json to tsv
        return dataframe.DataFrame(new_header_fields, new_data_fields) \
            .explode_json(internal_prefix, internal_prefix, excluded_cols = self.excluded_cols, nested_cols = self.nested_cols) \
            .remove_prefix(internal_prefix)



from omigo_core import dataframe, utils
from jira import JIRA
import os
import json

# env variables
JIRA_API_USER = "JIRA_API_USER"
JIRA_API_PASS = "JIRA_API_PASS" # nosec
JIRA_API_AUTH_TOKEN = "JIRA_API_AUTH_TOKEN" #nosec

# api handler for searching jira
class JiraSearch:
    def __init__(self, server = None, username = None, password = None, auth_token = None, proxies = None, verify = True):
        # warn
        utils.warn_once("JiraSearch: This is work in progress in extensions package. Some of the constants need to be decoupled")

        # validation
        if (server is None):
            raise Exception("JiraSearch: server is None")

        # init
        self.server = server
        self.jira_instance = None
        self.fields_mapping = {}

        # instantiate and return
        jira_options = {"server": self.server, "verify": verify, "headers": {'content-type': 'application/json'}}
        utils.info("JiraSearch: instantiating with options: {}".format(jira_options))

        # debug
        if (proxies is not None):
            utils.info("Using proxy: {}".format(proxies))

        # check for credentials
        if (username is not None and password is not None):
            self.jira_instance = JIRA(jira_options, basic_auth = (username, password), proxies = proxies)
        elif (auth_token is not None):
            self.jira_instance = JIRA(jira_options, token_auth = auth_token, proxies = proxies)
        else:
            raise Exception("JiraSearch: No valid authentication mechanism found")

        # build fields mapping
        for field in self.jira_instance.fields():
            field_name = str(field["name"])
            field_type = str(field["schema"]["type"]) if ("schema" in field and "type" in field["schema"]) else ""
            self.fields_mapping[str(field["id"])] = {"name": field_name, "type": field_type, "field": field}

    def get_server(self):
        return self.server

    def get_instance(self):
        return self.jira_instance

    def search_issues(self, query, max_results = None, datatype_map = None, dmsg = ""):
        dmsg = utils.extend_inherit_message(dmsg, "JiraSearch: search_issues")

        # resolve num_warnings
        max_results = int(utils.resolve_default_parameter("max_results", max_results, "10", "{}: {}".format(dmsg, query)))
        datatype_map = utils.resolve_default_parameter("datatype_map", datatype_map, {}, "{}: {}".format(dmsg, query))

        # the query is assumed to be resolved
        search_results = self.jira_instance.search_issues(query, maxResults = max_results)

        # check for empty
        if (search_results is None or len(search_results) == 0):
            return dataframe.create_empty()

        # iterate
        result_xdfs = []
        for search_result in search_results:
            # jira has a special attribute that holds all fields
            key = search_result.raw["key"]
            result_fields = search_result.raw["fields"]

            # trace
            for k in result_fields.keys():
                value = str(result_fields[k]) if (result_fields[k] is not None) else ""

            # iterate and add each available field
            mp = {}

            # assign key
            mp["key"] = str(key)

            # iterate on selected cols
            for k in result_fields.keys():
                # take value
                value = result_fields[k]

                # Ignore None as even dictionary can be None
                if (value is None):
                    value = ""
                    continue

                if (isinstance(value, (list)) and len(value) == 0):
                    value = ""
                    continue

                if (isinstance(value, (dict)) and len(value) == 0):
                    value = ""
                    continue

                # map the key
                k2 = k
                field_type = self.fields_mapping[k]["type"] if (k in self.fields_mapping) else ""

                # map the custom fields
                if (k.startswith("customfield_") and k in self.fields_mapping):
                    k2 = self.fields_mapping[k]["name"]

                # utils.trace("{}: Key: {}, Field: {}, Value: {}".format(dmsg, k2, field_type, str(value)[0:50].replace("\n", " ").replace("\r", " ")))
                # do json encoding if needed
                if (field_type in ("string")):
                    value_str = str(value)
                    if (value_str.startswith("{\"") and value_str.endswith("}")):
                        mp_value = json.dumps(json.loads(value_str))
                        mp["{}:json_encoded".format(k2)] = mp_value 
                    elif (value_str.startswith("[{\"") and value_str.endswith("}]")):
                        mp_value = json.dumps(json.loads(value_str))
                        mp["{}:json_encoded".format(k2)] = mp_value
                    elif (value_str.startswith("{'") and value_str.endswith("}")):
                        utils.warn_once("{}: JSON encoded string found with single quotes. Doing best effort transformation and parsing".format(dmsg))
                        mp_value = json.dumps(json.loads(value_str.replace("'", "\"")))
                        mp["{}:json_encoded".format(k2)] = mp_value 
                    elif (value_str.startswith("[{'") and value_str.endswith("}]")):
                        utils.warn_once("{}: JSON encoded string found with single quotes. Doing best effort transformation and parsing".format(dmsg))
                        mp_value = json.dumps(json.loads(value_str.replace("'", "\"")))
                        mp["{}:json_encoded".format(k2)] = mp_value
                    else:
                        mp[k2] = value_str
                elif (field_type in ("date", "datetime", "group", "number")):
                    utils.warn_once("{}: be careful while adding entries here. This can be a complex structure".format(dmsg))
                    mp[k2] = str(value)
                elif (isinstance(value, (dict))):
                    mp_value = json.dumps(value)
                    mp["{}:json_encoded".format(k2)] = mp_value 
                elif (field_type in ("array") and len(value) > 0):
                    if (isinstance(value[0], (dict))):
                        mp_value = json.dumps(value)
                        mp["{}:json_encoded".format(k2)] = mp_value
                    else: 
                        mp[k2] = ",".join(list([str(v) for v in value]))
                elif (isinstance(value, (str, int, float))):
                    mp[k2] = str(value)
                else:
                    mp_vars = vars(value)
                    mp_value = {}
                    for k in mp_vars:
                        if (k.startswith("_") == False):
                            mp_value[str(k)] = str(mp_vars[k])
                    mp["{}:json_encoded".format(k2)] = json.dumps(mp_value)

            # inner transformation function
            def __raw_json_custom_field_mapping__(raw_json):
                # base condition
                if (raw_json is None or isinstance(raw_json, (dict)) == False):
                    return raw_json

                # transform
                mp_new = {}
                for k in raw_json:
                    if (k.startswith("customfield_") and k in self.fields_mapping):
                        mp_new[self.fields_mapping[k]["name"]] = __raw_json_custom_field_mapping__(raw_json[k])
                    else:
                        mp_new[k] = __raw_json_custom_field_mapping__(raw_json[k])

                # return
                return mp_new

            # append raw along with custom field transformation
            mp["raw"] = json.dumps(__raw_json_custom_field_mapping__(search_result.raw))

            # append to tsvs
            result_xdfs.append(dataframe.from_maps([mp]))

        # merge
        result = dataframe.merge_union(result_xdfs)

        # return
        return result

# DF for search jira
class JiraDF(dataframe.DataFrame):
    def __init__(self, header, data_fields, jira_search = None, server = None, username = None, password = None, auth_token = None, verify = True):
        super().__init__(header, data_fields)

        # instantiate
        self.jira_search = jira_search if (jira_search is not None) else JiraSearch(server = server, username = username, password = password, auth_token = auth_token, verify = verify)

    def search_issues(self, query_template, prefix, extra_cols = None, max_results = 10, dmsg = ""):
        dmsg = utils.extend_inherit_message(dmsg, "JiraDF: search_issues")

        def __search_issues_explode_func__(mp):
            # resolve query
            query = utils.replace_template_props(mp, query_template)
            utils.info("JiraDF: search_issues: resolved query: {}".format(query))

            # call jira search
            results = self.jira_search.search_issues(query, extra_cols = extra_cols, max_results = max_results, dmsg = dmsg)

            # return
            return results.to_maps()

        # return
        return self \
            .explode(".*", lambda mp: __search_issues_explode_func__(mp), prefix, collapse = False, dmsg = dmsg)




from omigo_core import dataframe, utils
from omigo_ext import sql_helper
import os

from pyspark import SparkContext, SparkConf
from pyspark.sql import SQLContext
from pyspark.sql import SparkSession

class SparkClient(sql_helper.HadoopSqlBase):
    def __init__(self, props = None, master = "yarn", app_name = "test"):
        # TODO: This needs to be changed
        import findspark
        if ("SPARK_HOME" in os.environ.keys()):
            findspark.init()

        # create spark session
        spark_conf = SparkConf() \
            .setMaster(master) \
            .setAppName(app_name) \
            .set("spark.sql.hive.caseSensitiveInferenceMode", "NEVER_INFER") \
            .set("spark.dynamicAllocation.enabled", "false") \
            .set("spark.yarn.maxAppAttempts", "1") \
            .set("spark.driver.maxResultSize", "4G") \
            .set("spark.executor.memoryOverhead", "1G") \
            .set("spark.executor.instances", "10") \
            .set("spark.executor.cores","4") \
            .set("spark.sql.shuffle.partitions", "100") \
            .set("spark.submit.deployMode", "client")

        # set parameters
        if (props is not None):
            for k in props.keys():
                spark_conf = spark_conf.set(k, props[k])

        # instantiate sql context
        spark_context = SparkContext.getOrCreate(conf = spark_conf)

        # instantiate spark
        self.spark = SparkSession.builder.enableHiveSupport().getOrCreate()
        self.sql_context = SQLContext(spark_context)

    # Override
    def execute_query_in_engine(self, query):
        # Execute the query
        pandas_df = self.sql_context.sql(query)

        # convert to tsv
        xdf = dataframe.from_df(pandas_df.toPandas())

        # call stop
        self.__stop__()

        # return
        return xtsv.get_columns(), xdf.get_data_fields() 

    # stop
    def __stop__(self):
        utils.info("SparkClient: stop called")
        self.spark.stop()
        self.spark = None
        self.spark_context = None
        self.sql_context = None

